//========= Copyright  1996-2003, Valve LLC, All rights reserved. ============
//
// The copyright to the contents herein is the property of Valve, L.L.C.
// The contents may be used and/or copied only with the written permission of
// Valve, L.L.C., or in accordance with the terms and conditions stipulated in
// the agreement/contract under which the contents have been supplied.
//
// Purpose: 
//
// $NoKeywords: $
//=============================================================================

#ifndef DIALOGMANAGER_H
#define DIALOGMANAGER_H

#ifdef _WIN32
#pragma once
#endif

#include <UtlLinkedList.h>
#include <KeyValues.h>
#include <vgui_controls/PHandle.h>

namespace vgui
{

//-----------------------------------------------------------------------------
// Purpose: utility class, maps a set of ID's to dialogs
//			used to manage sets of similar dialogs (object property dialogs, etc.)
//-----------------------------------------------------------------------------
template <class TDialog>
class DialogManager
{
public:
	// new dialog factory function
	typedef TDialog *(*CreateNewDialogFunc_t)(int dialogID);

	// constructor
	DialogManager(CreateNewDialogFunc_t createDialogFunc);

	// finds the dialog by the specified ID
	TDialog *FindDialog( int dialogID, bool bCreate);

	// opens the dialog; creating it if specified
	void ActivateDialog(int dialogID, bool bCreate);

	// closes all the dialogs
	void CloseAll();

	// closes and deletes all the dialogs
	void CloseAndDeleteAll();

	void SetParent( vgui::VPANEL parent );

private:
	// checks if an index in the dialog list is valid; if it has been deleted, removes the entry
	bool ValidateIndex(int index);

	struct DialogItem_t
	{
		int id;
		DHANDLE<TDialog> dlg;
	};

	CUtlLinkedList<DialogItem_t, int> m_Dialogs;
	CreateNewDialogFunc_t m_CreateFunc;
	vgui::VPANEL m_pVGUIParentPanel;
};


// constructor
template <class TDialog>
inline DialogManager<TDialog>::DialogManager(CreateNewDialogFunc_t createDialogFunc)
{
	m_CreateFunc = createDialogFunc;
}

// finds the dialog; creating it if necessary
template <class TDialog>
inline TDialog *DialogManager<TDialog>::FindDialog(int dialogID, bool bCreate)
{
	for (int i = 0; i < m_Dialogs.MaxElementIndex(); i++)
	{
		if (ValidateIndex(i) && m_Dialogs[i].id == dialogID)
		{
			return m_Dialogs[i].dlg;
		}
	}

	if (bCreate)
	{
		int newIndex = m_Dialogs.AddToTail();
		if (m_CreateFunc)
		{
			m_Dialogs[newIndex].dlg = m_CreateFunc(dialogID);
		}
		else
		{
			m_Dialogs[newIndex].dlg = new TDialog(NULL, dialogID);
		}
		m_Dialogs[newIndex].dlg->SetParent( m_pVGUIParentPanel );

		m_Dialogs[newIndex].id = dialogID;
		return m_Dialogs[newIndex].dlg;
	}

	// dlg not found, not created
	return NULL;
}

// opens the dialog; creating it if necessary
template <class TDialog>
inline void DialogManager<TDialog>::ActivateDialog(int dialogID, bool bCreate)
{
	TDialog *dlg = FindDialog(dialogID, bCreate);
	if (dlg)
	{
		dlg->Activate();
	}
}

// closes all the dialogs
template <class TDialog>
inline void DialogManager<TDialog>::CloseAll()
{
	for (int i = 0; i < m_Dialogs.MaxElementIndex(); i++)
	{
		if (ValidateIndex(i))
		{
			m_Dialogs[i].dlg->PostMessage(m_Dialogs[i].dlg, new KeyValues("Close"));
		}
	}
}

// closes and deletes all the dialogs
template <class TDialog>
inline void DialogManager<TDialog>::CloseAndDeleteAll()
{
	CloseAll();
	for (int i = 0; i < m_Dialogs.MaxElementIndex(); i++)
	{
		if (ValidateIndex(i))
		{
			m_Dialogs[i].dlg->MarkForDeletion();
		}
	}
	m_Dialogs.RemoveAll();
}

// checks if a dialog is valid; if it has been deleted, removes the entry
template <class TDialog>
inline bool DialogManager<TDialog>::ValidateIndex(int index)
{
	if (m_Dialogs.IsValidIndex(index))
	{
		if (m_Dialogs[index].dlg.Get())
		{
			return true;
		}
		else
		{
			// entry has been deleted; removed
			m_Dialogs.Remove(index);
		}
	}
	return false;
}

template <class TDialog>
inline void DialogManager<TDialog>::SetParent( vgui::VPANEL parent )
{
	m_pVGUIParentPanel = parent;
}


} // namespace vgui

#endif // DIALOGMANAGER_H
